const LATEST_VIEW = 0;
const CHANGES_VIEW = 1;
const HISTORY_VIEW = 2;

function ModelTreeData(rdfservice, rdfcu, builder, view)
{
  this.rdfservice = rdfservice;
  this.rdfcu = rdfcu;
  this.builder = builder;
  this.view = view;
  // Set up some RDF resources...
  this.rr =
  {
    root: this.rdfservice.GetResource(
      "http://www.cellml.org/tools/pcenv/all-models"),
    name: this.rdfservice.GetResource(
      "http://www.cellml.org/tools/pcenv/models#Name"),
    open: this.rdfservice.GetResource(
      "http://home.netscape.com/NC-rdf#open"),
    true: this.rdfservice.GetLiteral("true")
  }

  this.persist = this.rdfservice.GetDataSource("rdf:local-store");

  var c = Components.classes["@mozilla.org/rdf/datasource;1?name=in-memory-" +
                             "datasource"];
  var ifce = Components.interfaces.nsIRDFDataSource;


  // Create three new datasources, one for each view...
  this.sources = [];
  this.rdfroot = [];

  this.nodeByResource = {};
  this.resourceByNode = {};
  this.latestResourceByTopNode = {};

  var i;
  for (i = 0; i < 3; i++)
  {
    this.sources[i] = c.createInstance(ifce);
    // Create a top-level container...
    this.rdfroot[i] = this.rdfcu.MakeSeq(this.sources[i], this.rr.root);
  }

  // It is now time to recursively set up the RDF sources from the context...
  var lmRoot = window.context.loadedModels;
  var sourceStack = [lmRoot];
  // We do a depth-first search, building the CHANGES_VIEW RDF, and
  // making lists in preparation for HISTORY_VIEW and LATEST_VIEW.
  var propStack = [[this.rr.root, this.rdfroot[CHANGES_VIEW], null]];
  this.rdfChangeSequenceByList = {};
  this.rdfChangeSequenceByList[
    lmRoot.QueryInterface(Components.interfaces.ICORBAWrapper).objid] =
    this.rdfroot[CHANGES_VIEW];
  var latestlist = [];
  var historylist = [];

  while (sourceStack.length)
  {
    var ssi = sourceStack.pop();
    var props = propStack.pop();
    var it = ssi.iterateModelNodes();

    while (true)
    {
      var n = it.nextModelNode();
      if (n == null)
        break;
      var l = n.derivedModels;
      sourceStack.push(l);
      var rsrc = this.rdfservice.GetAnonymousResource();
      this.persist.Assert(rsrc, this.rr.open, this.rr.true, true);
      this.nodeByResource[rsrc.ValueUTF8] = n;
      this.resourceByNode[n.QueryInterface(Components.interfaces.
                                           ICORBAWrapper).objid] = rsrc;
      var seq = this.rdfcu.MakeSeq(this.sources[CHANGES_VIEW], rsrc);
      this.rdfChangeSequenceByList[
        l.QueryInterface(Components.interfaces.ICORBAWrapper).objid] = seq;
      var latestdata = props[2];
      var mt = n.modificationTimestamp;
      if (latestdata == null)
      {
        // Note: Don't update latestdata[3], it is supposed to be stuck as the
        // toplevel node corresponding to the latest, not the latest.
        // This differs from latestdata[0], which gets updated.
        latestdata = [n, mt, rsrc, n];
        latestlist.push(latestdata);
      }
      else if (latestdata[1] < mt)
      {
        latestdata[0] = n;
        latestdata[1] = mt;
        latestdata[2] = rsrc;
      }
      // Add a record in the history list.
      historylist.push([n, mt, rsrc]);

      // Add the element...
      props[1].AppendElement(rsrc);
      // Set the name...
      var i;
      var l = this.rdfservice.GetLiteral(n.name);
      for (i = 0; i < 3; i++)
        this.sources[i].Assert(rsrc, this.rr.name, l, true);

      // Set up propstack for the next trip...
      propStack.push([rsrc, seq, latestdata]);
    }
  }
  // Sort the history list...
  historylist.sort(function(a,b) {return (a[1] - b[1]); });

  this.historySeq = this.rdfcu.MakeSeq(this.sources[HISTORY_VIEW],
                                       this.rr.root);
  for (record in historylist)
  {
    var record = historylist[record];
    var rsrc = record[2];
    this.historySeq.AppendElement(rsrc);
  }

  this.latestSeq = this.rdfcu.MakeSeq(this.sources[LATEST_VIEW],
                                      this.rr.root);
  for (record in latestlist)
  {
    var record = latestlist[record];
    var rsrc = record[2];
    this.latestSeq.AppendElement(rsrc);
    var objid = record[3].QueryInterface(Components.interfaces.ICORBAWrapper).
      objid;
    this.latestResourceByTopNode[objid] = rsrc;
  }

  this.initialiseModelRenamed();
  this.initialiseModelAdded();
  this.initialiseModelRemoved();
}

ModelTreeData.prototype =
{
  QueryInterface: function(aIID)
  {
    if (aIID.equals(Components.interfaces.cellml_context_IModelNodeMonitor) ||
        aIID.equals(Components.interfaces.cellml_context_IModelListMonitor) ||
        aIID.equals(Components.interfaces.nsISupports))
      return this;
    return NULL;
  },

  initialiseModelRenamed: function()
  {
    var obj = {};
    this.modelRenamedProxy = CreateUpdateUI(function() {
      var t = this.queue.shift();
      var ol = t.d.rdfservice.GetLiteral(t.oldName);
      var nl = t.d.rdfservice.GetLiteral(t.newName);
      var rsrc = t.d.getResourceByNode(t.renamedNode);
      if (rsrc == null)
        return;
      // The final bit must be done on the main thread...
      var i;
      for (i = 0; i < 3; i++)
        t.d.sources[i].Change(rsrc, t.d.rr.name, ol, nl);
    }, obj);
    this.modelRenamedObj = obj.value;
    this.modelRenamedObj.queue = [];
  },
  
  modelRenamed: function(renamedNode, newName)
  {
    var t = {};
    t.oldName = renamedNode.name;
    t.newName = newName;
    t.renamedNode = renamedNode;
    t.d = this;
    this.modelRenamedObj.queue.push(t);
    this.modelRenamedProxy.run();
  },

  initialiseModelAdded: function()
  {
    var obj = {};
    this.modelAddedProxy = CreateUpdateUI(function() {
      var t = this.queue.shift();
      var l = t.newNode.parentList;
      if (l == null)
        return;
      
      // We will always insert at a certain point in the change sequence...
      var changeseq = t.d.getChangeSequenceByList(l);
      if (changeseq == null)
        return;
      
      // The history list is very simple: we just add the new resource at the
      // end...
      var historyseq = t.d.historySeq;
      
      // Find the toplevel object.
      var tla = t.newNode, tlb;
      while ((tlb = tla.parentList.parentNode) != null)
      {
        tla = tlb;
      }
      // This might return null, if this is a completely new model,
      // rather than a new derivative model.
      var oldLatest = t.d.getLatestResourceByNode(tla);
      var latestseq = t.d.latestSeq;

      // Manufacture a new resource...
      var rsrc = t.d.rdfservice.GetAnonymousResource();

      // Make it open...
      t.d.persist.Assert(rsrc, t.d.rr.open, t.d.rr.true, true);

      // Update nodeByResource and resourceByNode...
      t.d.nodeByResource[rsrc.ValueUTF8] = t.newNode;
      t.d.resourceByNode[
        t.newNode.QueryInterface(Components.interfaces.ICORBAWrapper).objid] =
        rsrc;
      t.d.latestResourceByTopNode[
        tla.QueryInterface(Components.interfaces.ICORBAWrapper).objid] =
        rsrc;
      // Produce a new sequence for the changes view...
      var seq = t.d.rdfcu.MakeSeq(t.d.sources[CHANGES_VIEW], rsrc);
      var dn = t.newNode.derivedModels;
      var objid = dn.QueryInterface(Components.interfaces.
                                    ICORBAWrapper).objid;
      t.d.rdfChangeSequenceByList[objid] = seq;

      // Get the name literal of the new node...
      var name = t.d.rdfservice.GetLiteral(t.newNode.name);
    
      // Assert the name onto the new node...
      var i;
      for (i = 0; i < 3; i++)
        t.d.sources[i].Assert(rsrc, t.d.rr.name, name,
                              true);
    
      changeseq.AppendElement(rsrc);
      historyseq.AppendElement(rsrc);
      if (oldLatest != null)
      {
        var oldidx = latestseq.IndexOf(oldLatest);
        latestseq.RemoveElementAt(oldidx, false);
        latestseq.InsertElementAt(rsrc, oldidx, false);
      }
      else
        latestseq.AppendElement(rsrc);
      t.d.builder.rebuild();
      // I think it is useful to select the new model.
      var tb = t.d.builder.QueryInterface(Components.interfaces.nsIXULTreeBuilder);
      var idx = tb.getIndexOfResource(rsrc);
      t.d.view.selection.select(idx);
    }, obj);
    this.modelAddedObj = obj.value;
    this.modelAddedObj.queue = [];
  },

  modelAdded: function(newNode, depth)
  {
    var e = {};
    e.newNode = newNode;
    e.depth = depth;
    e.d = this;
    this.modelAddedObj.queue.push(e);
    this.modelAddedProxy.run();
  },

  findLatestChildButOne: function(root, excludeid)
  {
    var l = root.derivedModels;
    if (l == null)
      return null;
    var it = l.iterateModelNodes();
    var nmn;
    // Get our timestamp...
    var ts = root.modificationTimestamp;
    var hit = root;
    var objid = root.QueryInterface(Components.interfaces.
                                    ICORBAWrapper).objid;
    if (objid == excludeid)
    {
      return null;
    }
    
    var nmn;
    while ((nmn = it.nextModelNode()) != null)
    {
      var m = this.findLatestChildButOne(nmn, excludeid);
      if (m == null)
        continue;
      if (m[1] > ts)
      {
        hit = m[0];
        ts = m[1];
      }
    }
    return [hit, ts];
  },
  
  initialiseModelRemoved: function()
  {
    var obj = {};
    this.modelRemovedProxy = CreateUpdateUI(function() {
      var t = this.queue.shift();
      // Find the resource...
      var rsrc = t.d.getResourceByNode(t.oldNode);
      if (rsrc == null)
        return;
      
      // Find the parent sequence...
      var l = t.parentList;
      
      if (l == null)
        return;
      var changeseq = t.d.getChangeSequenceByList(l);
      if (changeseq == null)
        return;
      
      // The history list is a straight delete...
      var historyseq = t.d.historySeq;
      
      // There are two cases for the latest list...
      //  1) The node is not the latest. In that case do nothing.
      //  2) The node is the latest. Replace it with the next newest
      //     surviving node.
      var tla = t.oldNode, tlb;
      while (l != null)
      {
        tlb = l.parentNode;
        if (tlb == null)
          break;
        tla = tlb;
        l = tla.parentList;
      }
      // This might return null, if this is a completely new model,
      // rather than a new derivative model.
      var latestseq = t.d.latestSeq;
      var latest = t.d.getLatestResourceByNode(tla);
      // latest can't be null unless our internal state is messed up, so I
      // am deliberately not adding a check for it.
      var rLatest = latest.ValueUTF8;
      var rOld = rsrc.ValueUTF8;
      var newRsrc = null;
      var delRsrcLatest = false;
      if (rLatest == rOld)
      {
        delRsrcLatest = true;
        var objidOld = t.d.nodeByResource[rOld].
          QueryInterface(Components.interfaces.ICORBAWrapper).objid;
        var replacement = t.d.findLatestChildButOne(tla, objidOld);
        // Find the resource for the replacement...
        newRsrc = null;
        if (replacement != null)
        {
          newRsrc = t.d.getResourceByNode(replacement[0]);
        }
        
        t.d.latestResourceByTopNode[
          tla.QueryInterface(Components.interfaces.ICORBAWrapper).objid] =
          newRsrc;
      }

      // Remove from the resourceByNode array so it disappears later.
      t.d.resourceByNode[t.oldNode.QueryInterface(Components.interfaces.
                                                  ICORBAWrapper).objid] = null;
      t.d.nodeByResource[rsrc.ValueUTF8] = null;

      changeseq.RemoveElement(rsrc, true);
      historyseq.RemoveElement(rsrc, true);
      if (delRsrcLatest)
      {
        if (newRsrc)
        {
          var idx = latestseq.IndexOf(rsrc);
          latestseq.RemoveElementAt(idx, false);
          latestseq.InsertElementAt(newRsrc, idx, false);
        }
        else
          latestseq.RemoveElement(rsrc, true);
      }
      this.builder.rebuild();
    }, obj);
    this.modelRemovedObj = obj.value;
    this.modelRemovedObj.queue = [];
    this.modelRemovedObj.builder = this.builder;
  },

  modelRemoved: function(oldNode, depth)
  {
    // If necessary, recursively remove children first (this ensures that
    //   we correctly compute things like the new latest node, although it
    //   is more expensive than what is required).
    var it = oldNode.derivedModels.iterateModelNodes();
    var nmn;
    while ((nmn = it.nextModelNode()) != null)
      this.modelRemoved(nmn, depth + 1);

    var t = {oldNode: oldNode, d: this};
    t.parentList = oldNode.parentList;
    this.modelRemovedObj.queue.push(t);
    this.modelRemovedProxy.run();
  },

  getResourceByNode: function(node)
  {
    var objid = node.QueryInterface(Components.interfaces.ICORBAWrapper).objid;
    return this.resourceByNode[objid];
  },

  getChangeSequenceByList: function(list)
  {
    var objid = list.QueryInterface(Components.interfaces.ICORBAWrapper).objid;
    return this.rdfChangeSequenceByList[objid];
  },

  getLatestResourceByNode: function(node)
  {
    var objid = node.QueryInterface(Components.interfaces.ICORBAWrapper).
      objid;
    return this.latestResourceByTopNode[objid];
  }
}
