function ImportLoadCompleted(r, imp, model, importURL)
{
  if (IsTagged(model, "load_failed"))
    return;
  UntagWithName(model, "import_in_progress");
  var rt = r.responseText;

  if (imp.wasInstantiated)
  {
    return;
  }
  try
  {
    imp.instantiateFromText(rt);
  }
  catch (e)
  {
    Status("Import " + importURL + " parse failed. Aborting model load.");
    TagWithName(model, "load_failed");
    return;
  }
  Status("Import " + importURL + " completed. Looking for further " +
         "imports...");
  GetMoreImportsOrRegister(importURL, model);
}

function ImportLoadFailed(r, imp, model, importURL)
{
  Status("Load of import " + importURL + " failed. Model load cancelled.");
  UntagWithName(model, "import_in_progress");
  TagWithName(model, "load_failed");
}

function AttemptAbsoluteURL(newURL, topURL)
{
  if (newURL.indexOf("://") > 0)
  {
    return newURL;
  }
  var protIdx = topURL.indexOf("://");
  var protEnd;
  if (protIdx < 0)
    protEnd = 0;
  else
    protEnd = protIdx + 3;
  var pathIdx = topURL.indexOf("/", protEnd);

  // If no protocol and no path, we might as well return the original string...
  if (protIdx < 0 && pathIdx < 0)
  {
    return newURL;
  }

  if (newURL[0] == "/")
  {
    // newURL is /foo/bar/test.xml, topURL is e.g. http://www.example.org
    if (pathIdx < 0)
    {
      return topURL + newURL;
    }
    return topURL.substr(0, pathIdx) + newURL
  }
  
  if (pathIdx < 0)
    return topURL + "/" + newURL;

  var lastSlash = topURL.lastIndexOf("/");
  var cand = topURL.substr(0, lastSlash);

  // Now try stripping ../ off beginning of newURL...
  while (true)
  {
    if (newURL[0] != '.')
      break;
    if (newURL[1] == '/')
    {
      newURL = newURL.substr(2);
      continue;
    }
    if (newURL[1] != '.')
      break;
    if (newURL[2] != '/')
      break;
    lastSlash = cand.lastIndexOf("/");
    if (lastSlash == -1)
      return cand + newURL;
    newURL = newURL.substr(3);
    cand = cand.substr(0, lastSlash);
  }
  return cand + "/" + newURL;
}

function GetImportLoadCompleted(imp, model, URL)
{
  return function (event){ ImportLoadCompleted(event.target, imp, model, URL); };
}

function GetImportLoadFailed(imp, model, URL)
{
  return function (event){ ImportLoadFailed(event.target, imp, model, URL); };
}

function RecursivelyFindUninstantiatedImports(modelURL, model, importSet)
{
  var cii = importSet.iterateImports();
  var allDone = true;
  var imp;
  while ((imp = cii.nextImport()) != null)
  {
    if (!imp.wasInstantiated &&
        !IsTagged(imp, "import_in_progress"))
    {
      allDone = false;
      TagWithName(imp, "import_in_progress");
      var URL = imp.xlinkHref.asText;
      var r = new XMLHttpRequest();
      URL = AttemptAbsoluteURL(URL, modelURL);
      Status("Attempting to load URL " + URL);
      r.onload = GetImportLoadCompleted(imp, model, URL);
      r.onerror = GetImportLoadFailed(imp, model, URL);
      r.open("GET", URL, true);
      try
      {
        r.send(null);
      }
      catch (e)
      {
        ImportLoadFailed({target: r}, imp, modelURL, URL);
      }
    }
    else
    {
      // Import already loaded, so visit its children...
      var im = imp.importedModel;
      if (im != null)
      {
        allDone &= RecursivelyFindUninstantiatedImports(modelURL, model, im.imports);
      }
      else
      {
        allDone = false;
      }
    }
  }
  return allDone;
}

function GetMoreImportsOrRegister(modelURL, model)
{
  if (!RecursivelyFindUninstantiatedImports(modelURL, model, model.imports))
    return;

  var modelName = model.cmetaId;
  if (modelName == "")
    modelName = modelURL;
  Status("Context-side model object created. Registering...");
  var c = window.context;
  if (c == null)
  {
    c = GetContext();
  }
  var lm = c.loadedModels;
  var n = lm.makeNode(model);
  n.name = modelName;
  n.stampModifiedNow();
  lm.addModel(n);
  Status("Model " + modelName + " loaded.");
}

function ModelLoadCompleted(r, modelURL)
{
  Status("Model XML loaded, transferring to your CellML context...");
  var c = window.context;
  if (c == null)
  {
    c = GetContext();
  }

  var ml = c.cellmlBootstrap.modelLoader;
  var m;
  try
  {
    m = ml.createFromText(r.responseText);
  }
  catch (e)
  {
    Status("Model load failed: " + ml.lastErrorMessage);
    return;
  }

  m.base_uri.asText = modelURL;

  GetMoreImportsOrRegister(modelURL, m);
}

function ModelLoadError(event)
{
  Status("Model load failed.");
}

function OpenHTTPRequest(url)
{
  var r = new XMLHttpRequest();
  r.open("GET", url);
  r.onload = function(event){ ModelLoadCompleted(event.target, url);}
  r.onerror = ModelLoadError;
  try
  {
    r.send(null);
  }
  catch (e)
  {
    ModelLoadError({target: r});
  }
}
